#!/usr/bin/perl -w

# check_quota.pl - nagios plugin to check Linux filesystem quotas using repquota
#
# (c) 2011,2014 by Frederic Krueger / fkrueger-dev-checkquota@holics.at
#
# Licensed under Creative Commons Attribution-ShareAlike 3.0 Unported
# There is no warranty of any kind, explicit or implied, for anything this software does or does not do.
# 
# Updates for this piece of software could be available under the following URL:
#   http://dev.techno.holics.at/check_quota/


### uses
use strict;
use v5.8;
use Getopt::Std;
#use Data::Dumper;             # for debugging

use lib "/usr/lib/nagios/plugins";
use lib "/usr/lib64/nagios/plugins";
use lib "/srv/nagios/libexec";
use utils qw (%ERRORS);

## in case we don't have nagios utils.pm in one of the above directories
## simply comment out the above line with "use utils" in it, and uncomment the 5 following lines
#%ERRORS = ('UNKNOWN' ,-1,
#           'OK'      , 0,
#           'WARNING' , 1,
#           'CRITICAL', 2
#          );



###
### DONT EDIT BELOW HERE
###

our ($PROG_NAME, $PROG_VERSION, $PROG_EMAIL, $PROG_URL) = ("check_quota", "0.24", 'fkrueger-dev-checkquota@holics.at', 'http://dev.techno.holics.at/check_quota/');

our %cache = ();
our $cache_timeout = 300;

## func
sub usage
{
  my $msg = shift;
  $msg = "" if (!defined($msg));

  print "\nusage: check_quota.pl [[-u <username> [-m <mountpoint>] [-g <groupname> -m <mountpoint>]] [-Q] [-p] [-h]\n";
  print "\n";
  print "   -g     check quota for a specific group (specify 'all' for all groups)\n";
  print "   -u     check quota for a specific user (specify 'all' for all users)\n";
  print "   -Q     show information on users/groups without a set quota as well\n";
  print "   -m     check quota for a specific mountpoint (specify 'all' for all mps; which is the default)\n";
  print "   -p     add perfdata\n";
  print "   -h     this help\n";
  print "\n";
  print "This plugin allows you to check the linux filesystem quota for a specific user, group or all quota users.\n";
  print "\n";
  print "$PROG_NAME v$PROG_VERSION is licensed under Creative Commons Attribution-ShareAlike 3.0 Unported License.
There is no warranty of any kind, explicit or implied, for anything this software does or does not do.

The main page for this plugin can be found at: $PROG_URL

(c) 2011,2014 by Frederic Krueger / $PROG_EMAIL

";

  if ($msg ne "")
  {
    print "\n";
    print "Error: $msg\n";
    print "\n";
    print "\n";
  }
  exit $ERRORS{'UNKNOWN'};
} # end sub usage


sub convertintohumanreadable
{
  my $val = shift;
  $val = 0 if (!defined($val));
  my @k = ("b", "K", "M", "G", "T", "P");

  my $out = "";
  my $cnt = 0;
  while ($val > 1023)
  {
    $val /= 1024;
    $cnt++;
  }
  $out = sprintf ( (($cnt == 0)?"%i%s":"%.2f%s"), $val, $k[$cnt]);

  return ($out);
} # end sub convertintohumanreadable


## the following is added for later usage.
sub getinfo
{
  my $fname = shift;
  $fname = "" if (!defined($fname));

  my $rc = "";

  # 1. delete old cache files
  if (defined($cache{$fname}))
  {
    if ($cache{$fname}{'timestamp'} > time-$cache_timeout)
    {
      delete ($cache{$fname});
    }
  }

  # 2. read files not in cache
  if (! defined($cache{$fname}))
  {
    if (open (A, "<$fname"))
    {
      my $content = "";
      { local $/ = undef; $content = <A>; };
      close(A);

      if (defined ($content))
      {
        $cache{$fname} = { 'timestamp' => time, 'content' => $content };
      }
    }
    else
    {
      warn "getinfo($fname): Couldn't open fname '$fname': $!\n";
    }
  }

  # 3. return info from cache
  if (defined($cache{$fname}))
  {
    $rc = $cache{$fname}{'content'};
  }

  return ($rc);
} # end sub getinfo


sub mountpoint_exists
{
  my $mp = shift;
  my $searchie = shift;
  $mp = "" if (!defined($mp));
  $searchie = "" if (!defined($searchie));

  # remove trailing slash, if there
  $mp =~ s/\/$//g;
  # spaces are written as \040 in /proc/mounts!
  $mp =~ s/ /\\040/sg;

  my $rc = 0;
  my $mpinfo = getinfo ("/proc/mounts");
  if ((defined($mpinfo)) and ($mpinfo ne ""))
  {
    my @tmp = split /\n/, $mpinfo;
    for (my $i=0; ($i <= $#tmp) and ($rc == 0); $i++)
    {
      my $line = $tmp[$i];
      next if ($line !~ /$searchie/);

      my @t = split /\s+/, $line;

      if ($t[1] eq $mp)
      {
        $rc = 1;
      }
    } # end if mpinfofile is readable
  } # end if mountpoint-infofile exists

  return ($rc);
} # end sub mountpoint_exists





## init
use vars qw($opt_g $opt_u $opt_v $opt_h $opt_p $opt_m $opt_Q
            $username $groupname $mountpoint $quota_status $quota_inuse $quota_softlimit $quota_hardlimit
            %ERRORS @rqout @outlist
            $verb_err $out);

$username = "";
$groupname = "";

our $bin_repquota = "/usr/sbin/repquota";
if ((!defined($bin_repquota)) or (! -e $bin_repquota))
  { $bin_repquota = `which repquota 2>/dev/null`;	chomp ($bin_repquota); }

# Turn this to 1 to see reason for parameter errors (if any)
$verb_err     = 0;


## args
if ($#ARGV < 0)
{
  &usage;
}
else
{
  getopts('ag:u:m:pQh');
}


if (defined($opt_h))
{
  usage();
}


## and the actual arg parsing
our $addperfdata = 0;	# dont add perfdata by default
if (defined($opt_p))
{
  $addperfdata = 1;
}

our $shownoquotaalso = 0;	# only show users/groups with quotas by default
if (defined($opt_Q))
{
  $shownoquotaalso = 1;
}


if ( (defined($opt_u)) and ($opt_u ne "") )
{
  $username = $opt_u;
  $username =~ s/[^0-9_a-zA-Z]//isg;
}
# it's either user or group, never both.
elsif ( (defined($opt_g)) and ($opt_g ne "") )
{
  $groupname = $opt_g;
  $groupname =~ s/[^0-9_a-zA-Z]//isg;
}



if ( (defined($opt_m)) and ($opt_m ne "") )
{
  if (($opt_m eq "all") or (mountpoint_exists ($opt_m, "quota")))
  {
    $mountpoint = $opt_m;
  }
  else
  {
    usage ("Supplied mountpoint '$opt_m' doesn't exist / have quota.. Exitting!");
    exit($ERRORS{'UNKNOWN'});
  }
}

###
### main
###

# get the repquota info we want to check
my $cmd = "$bin_repquota -c -v -s ";
if ((defined($username)) and ($username ne ""))
{
  $cmd .= "-u ";
}
elsif ((defined($groupname)) and ($groupname ne ""))
{
  $cmd .= "-g ";
}
else
{
  usage();
}

if ((defined($mountpoint)) and ($mountpoint ne "") and ($mountpoint ne "all"))
{
  $cmd .= "$mountpoint ";
}
else
{
  $cmd .= "-a ";
}

my @rqout = split /\n/, `$cmd`;


# and parse with it
my $perfdata = "";

my $nummatchingentities = 0;    # users/groups

foreach my $line (@rqout)
{
  if (defined($opt_u))
  {
    next if (($username ne "") and ($username ne "all") and ($line !~ /$username/i));
  }
  elsif (defined($opt_g))
  {
    next if (($groupname ne "") and ($groupname ne "all") and ($line !~ /$groupname/i));
  }

  my @dat = split /\s+/, $line;
  next if (($#dat < 7) or ($dat[4] !~ /^\d+/));

  my ($quotaname, $quota_status, $quota_inuse, $quota_softlimit, $quota_hardlimit) = ( $dat[0], $dat[1], $dat[2], $dat[3], $dat[4] );


  # convert input \d+G by \d+M (but without the M)
  $quota_inuse = ($quota_inuse =~ /^(.*)[Gg]$/) ? $1*1024*1024*1024 : ($quota_inuse =~ /^(.*)[Mm]$/) ? $1*1024*1024 : ($quota_inuse =~ /^(.*)[Kk]$/) ? $1*1024 : $quota_inuse;
  $quota_softlimit = ($quota_softlimit =~ /^(.*)[Gg]$/) ? $1*1024*1024*1024 : ($quota_softlimit =~ /^(.*)[Mm]$/) ? $1*1024*1024 : ($quota_softlimit =~ /^(.*)[Kk]$/) ? $1*1024 : $quota_softlimit;
  $quota_hardlimit = ($quota_hardlimit =~ /^(.*)[Gg]$/) ? $1*1024*1024*1024 : ($quota_hardlimit =~ /^(.*)[Mm]$/) ? $1*1024*1024 : ($quota_hardlimit =~ /^(.*)[Kk]$/) ? $1*1024 : $quota_hardlimit;

  # hardlimit > 1 only shows up for users that actually have a quota (via edquota)
  if ((defined($quota_status)) and (defined($quota_hardlimit)) and (($shownoquotaalso) or ($quota_hardlimit > 1)))
  {
    # Obviously the quota_status is set by what _you_ defined via edquota for the specific user/group
    #
    # XXX Someone on the nagios exchange asked for another way of doing this, but has yet to reply on
    # how this could be done differently (except for, maybe, making +- CRITICAL instead of WARNING or the like)
    #
    # I have yet to find the need for such a construct in my work, so I won't implement until someone else does.
    #
    # Message me what you want and I might add it to the script. ;-)

    my $retcode = $ERRORS {'UNKNOWN'};
    if (($shownoquotaalso) and ($quota_hardlimit <= 0))
    {
      $quota_status = "OK";
    }
    else
    {
      $quota_status = "OK"		if ($quota_status eq "--");
      $quota_status = "WARNING"	if ($quota_status eq "-+");
      $quota_status = "WARNING"	if ($quota_status eq "+-");
      $quota_status = "CRITICAL"	if ($quota_status eq "++");
    }

    $retcode = $ERRORS{$quota_status};

    if ((defined($quotaname)) and ($quotaname ne ""))
    {
      my $inusevalue = convertintohumanreadable ($quota_inuse);
      my $maxvalue = ($quota_hardlimit > 0) ? convertintohumanreadable ($quota_hardlimit) : "";

      $perfdata .= ((defined($opt_u)) ? "user":"group") ."_$quotaname=$quota_inuse;" .(($quota_softlimit > 0) ? $quota_softlimit : ""). ";" .(($quota_hardlimit > 0) ? $quota_hardlimit : ""). ";0; ";
      push @outlist, [ $retcode, $quota_status, sprintf ("%s - %s%s (%s: %s)", $quota_status, $inusevalue, (($quota_hardlimit <= 0) ? "" : " / $maxvalue"), ((defined($opt_u)) ? "login":"group"), $quotaname) ];
      $nummatchingentities++;
    }
  }
}



my $last_exitcode = -1;
my $last_exitstatus = "UNKNOWN";
my $quotafound	= 0;
my $hadoutput = 0;
my $txtoutput = "";

my $firstlineprinted = 0;
my $firstline = "";
foreach my $out (@outlist)
{
  $quotafound = 1;  $hadoutput = 1;
  if ($last_exitcode < $$out[0]) { $last_exitcode = $$out[0]; $last_exitstatus = $$out[1]; }
  if (!$firstlineprinted)       # save first line for later, in case we need to have it separately
    { $firstline = $$out[2]; $firstlineprinted = 1; }
  else
    { $txtoutput .= "$$out[2]"; }
  $txtoutput .= "\n"; 
#  print "got exitcode $$out[0] (status: $$out[1]) with data '$$out[2]'\n";
}


if ($txtoutput eq "\n") { $txtoutput = ""; }

if ($nummatchingentities > 1)
{
  $txtoutput = "$last_exitstatus - for matching $nummatchingentities " .((defined($opt_u)) ? "users" : "groups").  (($addperfdata > 0) ? " | $perfdata" : "") ."\n" . $firstline . $txtoutput;
}
else
{
  if (($firstline ne "") and ($firstline ne "\n")) { $txtoutput = "$firstline\n" . $txtoutput; }
}


my $entfound = 1;		# provide entityfound = yes, if all users or groups were chosen
if (($username ne "all") and ($groupname ne "all"))
{
  $entfound = ((defined($opt_u)) ? (getpwnam ($username)) : (getgrnam ($groupname)));
  $entfound = 0 if (!defined($entfound));
}


if (! $hadoutput)
{
  if (defined($opt_u))
  {
    if ($entfound <= 0)
    {
      if ($username eq "all")
      {
        $txtoutput .= "UNKNOWN - No users could be found?!\n";
      }
      else
      {
        $txtoutput .= "UNKNOWN - No such user '$username'\n";
      }
    }
    else
    {
      if ($username eq "all")
      {
        $txtoutput .= "UNKNOWN - No quota for any user could be found?!\n";
      }
      else
      {
        $txtoutput .= "UNKNOWN - No quota for user '$username'\n";
      }
    }
  } # end if user
  elsif (defined($opt_g))
  {
    if ($entfound <= 0)
    {
      if ($groupname eq "all")
      {
        $txtoutput .= "UNKNOWN - No groups could be found.\n";
      }
      else
      {
        $txtoutput .= "UNKNOWN - No such group '$groupname'\n";
      }
    }
    else
    {
      if ($groupname eq "all")
      {
        $txtoutput .= "UNKNOWN - No quota for any group could be found?!\n";
      }
      else
      {
        $txtoutput .= "UNKNOWN - No quota for group '$groupname'\n";
      }
    }
  } # end if group
} # end if no userfound or no quotafound

print "$txtoutput";
exit $last_exitcode;


